/***************************************************************************
                          mnbrak.cpp  -  description
                             -------------------
    begin                : Wed Sep 25 2002
    copyright            : (C) 2002 by Robert Marti
    email                : robertmarti@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "mnbrak.h"
#include <math.h>
#include <iostream.h>
#include "caract/rmlib/rmutils/rmutils.h"
#include "caract/rmlib/rmutils/rmfuncs.h"

inline void shift(double& a, double& b, double& c, double& d)
{
  a=b; b=c; c=d;
}


Mnbrak::Mnbrak(){
}

Mnbrak::~Mnbrak(){
}

void Mnbrak::solve(double &ax,	double &bx,	double &cx,
		double &fa,	double &fb,	double &fc,	rmFunc1D &F)
{
	const double GOLD = 1.618034, GLIMIT=100.0, TINY=1.0e-20;
	double ulim,u,r,q,fu,dum,Fu,diff,fabs_diff,max;


	fa=F(ax);
	fb=F(bx);
	if (fb > fa) {
		shift(dum,ax,bx,dum);
		shift(dum,fb,fa,dum);
	}
	cx=(bx)+GOLD*(bx-ax);
	fc=F(cx);
	while (fb > fc) {
		r=(bx-ax)*(fb-fc);
		q=(bx-cx)*(fb-fa);
		diff=q-r;
		fabs_diff=fabs(diff);
		max= rmUtils::fmax(fabs_diff,TINY);
		u=(bx)-((bx-cx)*q-(bx-ax)*r)/
		  (2.0* rmUtils::sign(max,diff));
		ulim=(bx)+GLIMIT*(cx-bx);
		if ((bx-u)*(u-cx) > 0.0) {
			fu=F(u);
			if (fu < fc) {
				ax=(bx);
				bx=u;
				fa=(fb);
				fb=fu;
				return;
			} else if (fu > fb) {
				cx=u;
				fc=fu;
				return;
			}
			u=(cx)+GOLD*(cx-bx);
			fu=F(u);
		} else if ((cx-u)*(u-ulim) > 0.0) {
			fu=F(u);
			if (fu < fc) {
			  bx=cx;
			  cx=u;
			  u=cx+GOLD*(cx-bx);
			  fb=fc;
			  fc=fu;
			  fu=F(u);
			}
		} else if ((u-ulim)*(ulim-cx) >= 0.0) {
			u=ulim;
			fu=F(u);
		} else {
			u=(cx)+GOLD*(cx-bx);
			fu=F(u);
		}
		shift(ax,bx,cx,u);
		shift(fa,fb,fc,fu);
	}
}
