/***************************************************************************
                          powell.cpp  -  description
                             -------------------
    begin                : Mon Sep 23 2002
    copyright            : (C) 2002 by Robert Marti
    email                : robertmarti@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "powell.h"
#include <math.h>
#include "caract/rmlib/rmutils/rmutils.h"
#include "caract/rmlib/rmmatrix/rmmatrix2d.h"
#include "caract/rmlib/Numrec/linmin.h"

Powell::Powell(){
	ItMax = 200;
}

Powell::~Powell(){
}

int Powell::solve( rmVector<double>& p,
	       rmFuncs& f,
	       double tol,
	       bool use_f_to_converge )
{
		// set up identity matrix
    const int n = p.dimX;
    rmMatrix2d<double> xi(n,n);
    xi.clear(0);
    for (int i=0; i<n; ++i) xi(i,i) = 1;

		// call powell
    int iter;
    double fret;
    solve(p, xi, tol, iter, fret, f, use_f_to_converge);

		// return number of iterations
    return iter;
}

void Powell::solve( rmVector<double> & p,
	        rmMatrix2d<double>& xi,
	        double ftol,
	        int& iter,
	        double& fret,
					rmFuncs& f,
	        bool use_f_to_converge )
{
	const int n = p.dimX;

	int ibig;
	double del,fp,fptt,t;

	rmVector<double> pt(p), xit;
	fret=f(p);
	for (iter=1;;++iter) {


		rmVector<double> prev(p);

		fp=fret;
		ibig=0;
		del=0.0;
		for (int i=0;i<n;i++) {
			xit = xi.getCol(i);
			fptt=fret;
			LinMin::solve(p,xit,fret,f);
			if (fabs(fptt-fret) > del) {
				del=fabs(fptt-fret);
				ibig=i;
			}
		}


				//***************************
				// test for convergence
				//***************************
		if (use_f_to_converge)
		{
			if (2.0*fabs(fp-fret) <= ftol*(fabs(fp)+fabs(fret)))
			    return;
		}
		else
		{
			bool carry_on = false;
			for (int i=p.dimX-1; i; --i)

				if (2*fabs(p(i)-prev(i)) >
					       ftol*(fabs(p(i))+fabs(prev(i))))
				{
					carry_on = true;
					break;
				}
			if (!carry_on) return;
		}


		if (iter == ItMax)
		{
		    cerr << "NR_Powell: exceeding maximum iterations.\n";
		    abort();
		}
		rmVector<double> ptt(p);
		ptt=ptt*2; ptt=ptt-pt;
		xit=p; xit=xit-pt;
		pt=p;
		fptt=f(ptt);
		if (fptt < fp) {
			t=2.0*(fp-2.0*fret+fptt)*rmUtils::sqr(fp-fret-del) -
			  del*rmUtils::sqr(fp-fptt);
			if (t < 0.0) {
				LinMin::solve(p,xit,fret,f);
				xi.setCol(ibig,xi.getCol(n));
				xi.setCol(n, xit);
			}
		}
	}
}
