/***************************************************************************
                          powell.h  -  description
                             -------------------
    begin                : Mon Sep 23 2002
    copyright            : (C) 2002 by Robert Marti
    email                : robertmarti@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef POWELL_H
#define POWELL_H


/**Implementation of the Powell minimsation algorithm, based on the onein Numerical Recipes.
  *@author Robert Marti
  */

#include "caract/rmlib/rmmatrix/rmmatrix2d.h"
#include "caract/rmlib/rmutils/rmfuncs.h"

const double NR_PowellDefaultTol = 1e-5;

class Powell {
private:
	int  ItMax;

public:
	Powell();
	~Powell();

/** Powell minimisation algorithm
	Inputs:		
			p: the guess of the point which is the function minimum
			xi: A set of directions used in the search (usually
			    the identity matrix)
			ftol: relative tolerance used in the
			      convergence test for the algorithm.
			f: the multi-valued function.
			use_f_to_converge: If true, the function values,
					   f(x), will be monitored to judge
					   when the method has converged.
					   If false, the variables, x[i],
					   will be monitored instead.
	Outputs:	p: the point which was located as the minimum.
			iter: number of iterations required by the algorithm
			fret: return the function value at exit
			      value of the point p i.e. the value
			      of the function at the minimum.
**/
	void solve( rmVector<double> &p,
	        rmMatrix2d<double>& xi,
	        double ftol,
	        int& iter,
	        double& fret,
					rmFuncs& f,
	        bool use_f_to_converge = true );

/** Simpler powell function where xi is identity
	Inputs:		p: the guess of the point which is the function minimum
			f: the multi-valued function.
			tol: relative tolerance used in the
			     convergence test for the algorithm.
			use_f_to_converge: If true, the function values,
					   f(x), will be monitored to judge
					   when the method has converged.
					   If false, the variables, x[i],
					   will be monitored instead.
	Outputs:	p: the point which was located as the minimum.
*/
	int solve( rmVector<double>& p,
	       rmFuncs& f,
	       double tol = NR_PowellDefaultTol,
	       bool use_f_to_converge = true );
};
#endif
