// rmDICOM format reading files
#include "rmdicom.h"
//#include "rmimage.h"
#include <string.h>

rmDICOM::rmDICOM()
{
  doPrint = FALSE ; //TRUE ;//FALSE; //TRUE;
  sliceSpacing = -1;
}


void rmDICOM::readFile(char * filename,int &dX, int &dY, int &bits, int &MVal,
		rmMatrix2d<imdata> &data, double * vs, double ss)
{
   FILE * fd;

   cout << "Opening File : "<< filename << endl;
   fd = fopen (filename, "r");
   if (fd == NULL) { cout << "DICOM: file opening error" << endl; exit(-1);}
   
   setEndian();  // set little or Big endian machine
   
	 if (readHeader(fd)== DCM_ERROR) {
	 cout << "DICOM: header reading error" << endl; exit(-1);}
	 readInfo(fd,data);
   dX= dimX;
   dY= dimY;
   bits = bpp;
   MVal = MaxValue;
   vs[0] = voxelSize[0];
   vs[1] = voxelSize[1];
   vs[2] = voxelSize[2];
   ss = sliceSpacing;
   cout << "END"<< endl; 
   fclose(fd);
}

void rmDICOM::readInfo (FILE * fd,rmMatrix2d<imdata>  &data)
{
   int group, element, len, res;
   dicom_types tt;
   char *svr;
//	char * value; unused
   int bits;

  finished = FALSE;

   do{
	group = element = len = 69;
	group = read16(fd);
	element = read16(fd);
	res = len= read16(fd);
	//if (lE) {
	//	svr = (char*) & res;	// retrieving vr if any
	//}
	//else {
	//printf("group (%04x, %04x)REs %04x \n",group,element,res);
	svr = new char[2];
	svr[1] = (res & 0x0000ff00) >> 8; 
	svr[0] = (res & 0x000000ff); 
	//}
	bits=explicitVR(svr,tt);
	if (explVR == TRUE) // There is a VR
	{
	    if (bits==16)  // no need to read anything else
	    {
		len = read16(fd);
	    }
	    if (bits==32) // need to read 16 bits more
	    {
		read16(fd); // following 16 bits are reserved (00000)
		len = read32(fd);
	    }
	}
	else 
	{
	    if (doPrint) cout << "No vr" << endl;
	}
	if (len>0) {
		checkGroup(group,element,len,svr,tt,fd,data);
	}
   }while (!feof(fd)&& !finished);
}

void rmDICOM::checkGroup (int group, int element, int len, char * svr,
		dicom_types tt, FILE * fd,rmMatrix2d<imdata> &data)
{
   char *info;
   int readValue = TRUE;

   char * buf, *buf2;
   int intvalue,i,j;
   float floatvalue;
   double doublevalue;
   unsigned short ushortvalue;
   unsigned long ulongvalue;
   short shortvalue;
   long longvalue;

   info="Unknown";
   switch (group) {
   case 0x0002 : 
		switch (element) {
	  	case 0x0000 :  info = "file meta elements group len"; break;
	   	case 0x0001 :  info = "file meta info version";       break;
      case 0x0002 :  info = "media storage SOP class uid";  break;
      case 0x0003 :  info = "media storage SOP inst uid";   break;
      case 0x0010 :  info = "transfer syntax uid";          break;
      case 0x0012 :  info = "implementation class uid";     break;
      case 0x0013 :  info = "implementation version name";  break;
      case 0x0016 :  info = "source app entity title";      break;
      case 0x0100:  info = "private info creator uid";     break;
      case 0x0102:  info = "private info";                 break;
   	}
	break;
   case 0x0008 :
        switch (element)  {
           case 0x00 :  info = "identifying group";   
		 	// len is ok;
			len=read16(fd);
			//cout << "VALUE " << read32(fd);	
			read32(fd);
			break;
           case 0x01 :  info = "length to end";                 break;
           case 0x05 :  info = "specific character set";  	break;
           case 0x08 :  info = "image type";              	break;
           case 0x10 :  info = "recognition code";              break;
           case 0x16 :  info = "SOP Class UID";           	break;
           case 0x18 :  info = "SOP Instance UID";              break;
           case 0x20 :  info = "study date";                    break;
           case 0x21 :  info = "series date";                   break;
           case 0x22 :  info = "acquisition date";              break;
           case 0x23 :  info = "image date";                    break;
           case 0x30 :  info = "study time";                    break;
           case 0x31 :  info = "series time";                   break;
           case 0x32 :  info = "acquisition time";              break;
           case 0x33 :  info = "image time";                    break;
           case 0x40 :  info = "data set type";                 break;
           case 0x41 :  info = "data set subtype";              break;
           case 0x50 :  info = "accession number";              break;
           case 0x60 :  info = "modality";                	break;
           case 0x70 :  info = "manufacturer";                  break;
           case 0x80 :  info = "institution name";        	break;
           case 0x90 :  info = "referring physician's name";    break;
           case 0x1010: info = "station name";                  break;
           case 0x103e: info = "series description";            break;
           case 0x1030: info = "study description";             break;
           case 0x1040: info = "institutional dept. name";      break;
           case 0x1050: info = "performing physician's name"; 	break;
           case 0x1060: info = "name phys(s) read stdy";        break;
           case 0x1070: info = "operator's name";         	break;
           case 0x1080: info = "admitting diagnoses description";break;
           case 0x1090: info = "manufacturer's model name";     break;
           case 0x1140: info = "referenced image sequence";	break;
        } 
	break;
   case 0x0010 :
        switch (element)  {
           case 0x00 :  info = "patient group";                 break;
           case 0x10 :  info = "patient name";            break;
           case 0x20 :  info = "patient ID";              break;
           case 0x30 :  info = "patient birthdate";       break;
           case 0x40 :  info = "patient sex";             break;
           case 0x1010: info = "patient age";             break;
           case 0x1030: info = "patient weight";          break;
           case 0x21b0: info = "additional patient history";break;
       }
       break;

   case 0x0018 :
       switch (element)  {
           case 0x00 :  info = "acquisition group";       break;
           case 0x10 :  info = "contrast/bolus agent";    break;
           case 0x15 :  info = "body part examined";      break;
           case 0x20 :  info = "scanning sequence";       break;
           case 0x21 :  info = "Sequence Variant";        break;
           case 0x22 :  info = "Scan Options";            break;
           case 0x23 :  info = "MR Acquisition Type";     break;
           case 0x24 :  info = "Sequence Name";           break;
           case 0x25 :  info = "Angio Flag";              break;
           case 0x30 :  info = "radionuclide";            break;
           case 0x50 :  info = "slice thickness";         
			    buf = (char *) malloc(len+1);
	   			if (fread(buf,len,1,fd)!=1) { 
					cout << "DICOM: Error reading value" << endl;
			   	}	
			   	buf[len]= '\0';
				voxelSize[2]= atof(buf);	
				readValue = FALSE;
		   	break;
           case 0x60:   info = "kvp";                     break;
           case 0x80 :  info = "repetition time";         break;
           case 0x81 :  info = "echo time";               break;
           case 0x82 :  info = "inversion time";          break;
           case 0x83 :  info = "Number of Averages";      break;
           case 0x84 :  info = "Imaging Frequency";       break;
           case 0x85 :  info = "Imaged Nucleus";          break;
           case 0x86 :  info = "Echo Number";             break;
           case 0x87 :  info = "Magnetic Field Strength"; break;
           case 0x88 :  info = "Spacing Between Slices";  
		   		buf = (char *) malloc(len+1);
	   			if (fread(buf,len,1,fd)!=1) { 
					cout << "DICOM: Error reading value" << endl;
			   	}	
			   	buf[len]= '\0';
				sliceSpacing= atof(buf);	
				readValue = FALSE;
		   	break;
           case 0x91 :  info = "Echo Train Length";       break;
           case 0x93 :  info = "Percent Sampling";        break;
           case 0x94 :  info = "Percent Phase Field of View"; break;
           case 0x95 :  info = "Pixel Bandwidth";         break;
           case 0x1020: info = "software version(s)";     break;
           case 0x1030: info = "protocol name";           break;
           case 0x1040: info = "Contrast/Bolus Route";    break;
           case 0x1062: info = "Nominal Interval";        break;
           case 0x1088: info = "Heart Rate";              break;
           case 0x1090: info = "Cardiac Number of Images"; break;
           case 0x1094: info = "Trigger Window";          break;
           case 0x1100: info = "Reconstruction Diameter"; break;
           case 0x1120: info = "gantry/detector tilt";    break;
           case 0x1150: info = "exposure time";           break;
           case 0x1151: info = "x-ray tube current";      break;
           case 0x1210: info = "convolution kernel";      break;
           case 0x1250: info = "Receiving Coil";          break;
           case 0x1251: info = "Transmitting Coil";       break;
           case 0x1310: info = "Acquisition Matrix";      break;
           case 0x1314: info = "Flip Angle";              break;
           case 0x1315: info = "Variable Flip Angle Flag"; break;
           case 0x1316: info = "SAR";                     break;
           case 0x1400: info = "Acquisition Device Processing Description"; 
			break;
           case 0x1401: info = "Acquisition Device Processing Code";
                        break;
           case 0x5100: info = "Patient Position";        break;
           case 0x5101: info = "View Position";           break;
           case 0x6000: info = "Sensitivity";             break;
       }
       break;
   case 0x0020 :
       switch (element)  {
           case 0x00 :  info = "relationship group";       break;
           case 0x0d :  info = "Study Instance UID";       break;
           case 0x0e :  info = "Series Instance UID";      break;
           case 0x10 :  info = "study id";                 break;
           case 0x11 :  info = "series number";            break;
           case 0x12 :  info = "acquisition number";       break;
           case 0x13 :  info = "image number";             break;
           case 0x20 :  info = "patient orientation";      break;
           case 0x30 :  info = "image position (ret)";     break;
           case 0x32 :  info = "Image Position (Patient)";
                                                           break;
           case 0x35 :  info = "image orientation (ret)";  break;
           case 0x37 :  info = "Image Orientation (Patient)";
                                                           break;
           case 0x50 :  info = "location (ret)";           break;
           case 0x52 :  info = "Frame of Reference UID";   break;
           case 0x60 :  info = "Laterality";               break;
           case 0x1002: info = "images in acquisition";    break;
           case 0x1040: info = "position reference";       break;
           case 0x1041: info = "slice location";           break;
           case 0x3401: info = "modifying device id";      break;
           case 0x3402: info = "modified image id";        break;
           case 0x3403: info = "modified image date";      break;
           case 0x3404: info = "modifying device mfg.";    break;
           case 0x3405: info = "modified image time";      break;
           case 0x3406: info = "modified image desc.";     break;
           case 0x4000: info = "image comments";           break;
           case 0x5000: info = "original image id";        break;
       } 
       break;
   case 0x0028 :
       switch (element)  {
           case 0x00 :  info = "image presentation group";      break;
           case 0x02 :  info = "samples per pixel";             break;
	   case 0x04 : info = "photometric interpretation"; break;
           case 0x05 :  info = "image dimensions (ret)";        break;
           case 0x06 :  info = "planar configuration";          break;
           case 0x08 :  info = "number of frames";              break;
           case 0x09 :  info = "frame increment pointer";       break;
           case 0x10 :  info = "rows";
			if (len==2) dimY =readu16(fd); 
			cout << "dim Y "<< dimY; 
			readValue = FALSE;
			break;
           case 0x11 :  info = "columns";
			if (len==2) dimX = readu16(fd);
			cout << " dim X "<< dimX <<endl;
			readValue = FALSE;
			break;
           case 0x30 :  info = "pixel spacing";                 
		   		buf = (char *) malloc(len+1);
	   			if (fread(buf,len,1,fd)!=1) { 
					cout << "DICOM: Error reading value" << endl;
			   	}	
			   	buf[len]= '\0';
				//i=strchr(buf,'\\');
				buf2 = strtok(buf,"\\");
				voxelSize[0]= atof(buf2);
				buf2 = strtok(NULL,"\\");
				voxelSize[1]= atof(buf2);	
				//cout <<"double values" << voxelSize[0] << voxelSize[1]<< endl;
				readValue = FALSE;
		   	break;
           case 0x31 :  info = "zoom factor";                   break;
           case 0x32 :  info = "zoom center";                   break;
           case 0x34 :  info = "pixel aspect ratio";            break;
           case 0x40 :  info = "image format (ret)";            break;
           case 0x50 :  info = "manipulated image (ret)";       break;
           case 0x51 :  info = "corrected image";               break;
           case 0x60 :  info = "compression code (ret)";        break;
           case 0x0100: info = "bits allocated";
			if (len==2) bpp= readu16(fd);
                        cout << "bits"<< bpp; 
                        readValue = FALSE;
			break;
           case 0x0101: info = "bits stored";
			if (len==2) MaxValue= 1<<readu16(fd);
                        cout << "MaxValue"<< MaxValue; 
                        readValue = FALSE;
			break;
           case 0x0102: info = "high bit"; break;
           case 0x0103: info = "pixel representation"; break;
	   case 0x0104: info = "Smallest Pixel Value"; break;
	   case 0x0105: info = "Largest Pixel Value";
                                 break;
           case 0x0200: info = "image location (ret)";          break;
           case 0x1050: info = "window center";
                                 break;
           case 0x1051: info = "window width";
                                 break;
           case 0x1052: info = "rescale intercept";
                                 break;
           case 0x1053: info = "rescale slope";
                                 break;
           case 0x1100: info = "gray lookup table desc (ret)";  break;
           case 0x1200: info = "gray lookup table data (ret)";  break;
           case 0x1201: info = "red table";
                                 break;
           case 0x1202: info="green table";
                                 break;
           case 0x1203: info="blue table";
                                 break;
       }
       break;

    case 0x4000 :
       info = "text";
       break;

    case 0x7FE0 :
       switch (element)  {
           case 0x00 :  info = "pixel data";
                        break;
           case 0x10 :  info = "pixel data";
						data.resize(dimX,dimY);
        		for (j=0;j<(int)dimY;j++)
            	for(i=0;i<(int)dimX;i++)
              {
								data(i,j) =readPixel(fd);
              }
							readValue = FALSE;
							finished= TRUE;
              break;
       }
       break;
    default :
       if (group >= 0x6000 && group <= 0x601e && (group & 1) == 0)
                    info = "overlay";
       //if (element == 0x0000)  info = "group length";
       if (element == 0x4000)  info = "comments";
       break;
   }//end switch group
   if (readValue==TRUE) {
   if (doPrint) {
       printf ("(%04x %04x): [%c%c] len:%i %26s",group,element,
                svr[0],svr[1],len,info);
       cout << "\n\t";
   }
   switch (tt) {
        case TSTRING:
	   buf = (char *) malloc(len+1);
	   if (fread(buf,len,1,fd)!=1) { 
		cout << "DICOM: Error reading value" << endl;
		//exit(-1);
	   }
	   buf[len]= '\0';
	   if (doPrint) cout << "str Value:" << buf;
	   break;
	case TSTRINT:
	   for (i=0;i<len/2;i++) {
		intvalue = read16(fd);
		if (doPrint) cout << "intValue: "<< intvalue;
	   }
	   if (doPrint) cout << endl;
	   break;
	case TFLOAT:
           if (fread(&floatvalue,len,1,fd)!=1) { 
                cout << "DICOM: Error reading value" << endl;
                //exit(-1);
           }
           if (doPrint) cout << "float Value:" << floatvalue; 
           break;
	case TDOUBLE:
	   if (fread(&doublevalue,len,1,fd)!=1) { 
                cout << "DICOM: Error reading value" << endl;
                //exit(-1);
           }
           if (doPrint) cout << "double Value:" << doublevalue;
           break;   
	case TSEQ: cout << "don't know how to handle SEQ"<< endl;break;
	case TNONE: if (doPrint) cout << "NONE???" << endl; break;
	case TLONG:
	   if (doPrint) cout << "long Value:";
	   for (i=0;i<len/4;i++) {
	   	longvalue = read32(fd);
           	if (doPrint) cout << longvalue << " ";
	   }
           break;
	case TSHORT:
	   if (doPrint) cout << "short Value:";
	   for (i=0;i<len/2;i++) {
	   	shortvalue= read16(fd);
           	if (doPrint) cout << shortvalue << " ";
	   }
           break;
	case TULONG:
           if (doPrint) cout << "ulong Value:";
           for (i=0;i<len/4;i++) {
	   	ulongvalue= readu32(fd);
           	if (doPrint) cout << ulongvalue << " ";
	   }
           break;
	case TUSHORT:
	   for (i=0;i<len/2;i++) {
	   	ushortvalue= readu16(fd);
           	if (doPrint) cout << ushortvalue << " ";
	   }
           break;
   }
   if (doPrint) cout << endl;
   } // endif

}

imdata rmDICOM::readPixel (FILE *fd)
{
    imdata value;
    uchar t1,t2;
    if (bpp>8) {
	fread(&t1, sizeof t1,1,fd);
   	fread(&t2, sizeof t2,1,fd);
	value= t1+t2*256;
    }

    if (bpp<=8) {
    	value= (imdata) getc(fd);
   }
   return value;
}
int rmDICOM::explicitVR(char * vr, dicom_types & tt)
{

   int bits=0;

   explVR = FALSE;
   tt = TNONE;

   switch (vr[0]) {
   case 'A':
	switch (vr[1]) {
	case 'E': tt = TSTRING;
		  explVR = TRUE;
		  bits= 16;
		break;  
	case 'S': tt = TSTRING;
                  explVR = TRUE;
		  bits= 16;
		break;
  	case 'T': tt = TSTRINT;
		  explVR = TRUE;
		  bits= 16;
		break;
	}
	break;
   case 'C':
	switch (vr[1]) {
	case 'S':
		tt = TSTRING;
		bits= 16;
		explVR = TRUE;
		break;
	}
	break;
   case 'D':
	switch (vr[1]) {
	case 'A':
		tt= TSTRING; 
		explVR = TRUE;
		bits= 16;
		break;
	case 'S':
		tt= TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
  	case 'T': 
		tt= TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
	}
	break;
   case 'F':
	switch (vr[1]) {
	case 'L': tt = TFLOAT;
		explVR = TRUE;
		bits= 16;
		break; 
	case 'D': tt = TDOUBLE;
		explVR = TRUE;
		bits= 16;
		break;
	}
	break;
   case 'I':
	switch (vr[1]) {
  	case 'S': tt = TSTRING;
		explVR = TRUE;
		bits = 16;
		break;
	}
	break;
   case 'L':
	switch (vr[1]) {
	case 'O': tt = TSTRING; 
		explVR = TRUE;
		bits= 16;
		break;
  	case 'T': tt = TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
	}
	break;
   case 'P':
	switch (vr[1]) {
  	case 'N': tt = TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
	}
	break;
   case 'S':
	switch (vr[1]) {
	case 'H': tt = TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
	case 'L': tt = TLONG;
		explVR = TRUE;
		bits= 16;
		break; 
	case 'S': tt = TSHORT;
		explVR = TRUE;
		bits= 16;
		break;
  	case 'T': tt = TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
	case 'Q': // it is a 32 bits
		tt =  TSEQ;
		explVR=TRUE;
		bits= 32;
		break;
	}
	break;
   case 'T':
	switch (vr[1]) {
	case 'M':
		tt = TSTRING; 
		explVR = TRUE;
		bits= 16;
		break;
	}
	break;
   case 'U':
	switch (vr[1]) {
	case 'I': tt = TSTRING; 
		explVR = TRUE;
		bits= 16;
		break;
	case 'L': tt = TULONG;
		explVR = TRUE;
		bits= 16;
		break;
  	case 'S': tt = TUSHORT;
		explVR = TRUE;
		bits= 16;
		break;
	case 'N': tt = TSTRING;
		explVR = TRUE;
		bits= 32;
		break;
	case 'T': tt = TSTRING;
		explVR = TRUE;
		bits= 16;
		break;
	}
	break;
   case 'O':
	switch (vr[1]) {
	case 'B':// 32 bits length
		tt = TSTRING;
                explVR = TRUE;
                bits= 32;
                break;
 
	case 'W': // 32 bits length
		tt = TSTRINT;
		explVR = TRUE;
		bits= 32;
		break;
	}
	break;
   }
   return bits;
}

int rmDICOM::readHeader(FILE * fd)
{
   char c;
   char dicm[4];
   int done;
  
   prused = FALSE;
   done = FALSE;
   do{
   	do { 
  	c = (char) getc(fd);
	if ((c!=0) && (c!='D') && (prused==FALSE)) prused= TRUE; 
   	}while ((c != 'D') && !feof(fd)); // possible starting of preamble
	if (fread(dicm,3,1,fd)==0) return DCM_ERROR;
	dicm[3]='\0';
        if (strcmp(dicm,"ICM")==0) {
	   cout << "DICM found" << endl;
	   done = TRUE;
	}
   } while ((done==FALSE) && !feof(fd));
   if (done==FALSE) return DCM_ERROR;
   else return DCM_OK;
}


void rmDICOM::setEndian()
{
   int l=1;
 
   uchar * u = (uchar*)& l;
   lE = u[0];
   cout << "Little Endian= "<< lE << endl;
}


int rmDICOM::read16( FILE * fd)
{
   uchar t1,t2;

   fread(&t1, sizeof t1,1,fd);
   fread(&t2, sizeof t2,1,fd);

   //if (lE) return t1+t2*256;
   //else    return t1*256+t2;
   return t1+t2*256;
}

unsigned int rmDICOM::readu16( FILE * fd)
{
   uchar t1,t2;

   fread(&t1, sizeof t1,1,fd);
   fread(&t2, sizeof t2,1,fd);

   //if (lE) return t1+t2*256;
   //else    return t1*256+t2;
   return t1+t2*256;
}

long rmDICOM::read32 (FILE * fd)
{
   uchar t1,t2,t3,t4;

   fread(&t1, sizeof t1,1,fd);
   fread(&t2, sizeof t2,1,fd);
   fread(&t3, sizeof t3,1,fd);
   fread(&t4, sizeof t4,1,fd);

   //if (lE)   return t1 + t2*256 + t3*256*256 + t4*256*256*256;
   //else     return t1*256*256*256 + t2*256*256 + t3*256 + t4;
   return t1 + t2*256 + t3*256*256 + t4*256*256*256;
}

unsigned long rmDICOM::readu32 (FILE * fd)
{
   uchar t1,t2,t3,t4;

   fread(&t1, sizeof t1,1,fd);
   fread(&t2, sizeof t2,1,fd);
   fread(&t3, sizeof t3,1,fd);
   fread(&t4, sizeof t4,1,fd);

   return t1 + t2*256 + t3*256*256 + t4*256*256*256;
}

void rmDICOM::Print(int i)
{
   doPrint= i;
}

