// 3D Image Class
//    Robert Marti
//	May 2001 

//#include "Rad/LinFilt.h"
#include "rmimage3d.h"
//#include "rmhisto1d.h"
#include "caract/rmlib/rmutils/rmutils.h"
#include <libgen.h>
#include <stdio.h>


#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>

int pstrcmp (const void *p1, const void *p2);

rmImage3d::rmImage3d()
{
	MaxValue=256;
	bits=8;
	dimX=dimY=dimZ=0;
	vsX=vsY=vsZ=1;
	origin[0]=origin[1]=origin[2]=0;
	sliceSpacing=0;
}

rmImage3d::rmImage3d(int dx,int dy, int dz, int vx, int vy, int vz)
{
	MaxValue=256;
	bits=8;
	dimX=dimY=dimZ=0;
	vsX=vsY=vsZ=1;
	sliceSpacing=0;
	origin[0]=origin[1]=origin[2]=0;
	resize(dx,dy,dz,vx,vy,vz);
}

rmImage3d::rmImage3d(int dx,int dy, int dz)
{
	MaxValue=256;
	bits=8;
	dimX=dimY=dimZ=0;
	vsX=vsY=vsZ=1;
	sliceSpacing=0;
	origin[0]=origin[1]=origin[2]=0;
	resize(dx,dy,dz);
}

const imdata3d & rmImage3d::operator() (int x, int y, int z) const
{
	return data(x,y,z);
}

imdata3d & rmImage3d::operator() (int x, int y, int z)
{
	return data(x,y,z);
}


/////////////////////
// read Image from a set of slices
//  	file: array of filenames
//  	start: starting slice in the array
//  	end: ending slice in the array (start..end-1 slices)
//  	norm: boolean, do slice normalisation while opening.
//		side: side to cut patientwise left or right (L or R) N for whole image
//////////////////////
void rmImage3d::readImage (char **file, int start, int end, int norm,
	char side)
{
	int i,reduce;
 	rmImage slice;
 
 	reduce = FALSE;
 	slice.read(file[start]);
/*	if (slice.bits>8) 
	{
		slice.RBits(8); // reducing bit size
		reduce = TRUE;
	}
	*/
	bits = slice.bits;
	MaxValue = slice.MaxValue;
	cout << "Bits: " << bits << " MaxValue: "<< MaxValue << endl;
 	if (norm) slice.norm();

	switch (side){
	case 'R': cout << "Cutting Right breast" << endl;
			slice.cut(0,0,(slice.dimX/2)-1,slice.dimY-1);break;
	case 'L': cout << "Cutting Left breast" << endl;
			slice.cut(slice.dimX/2,0,slice.dimX-1,slice.dimY-1);
			break;
	}
 //slice.Save("cut.byte");
 	vsX = slice.voxelSize[0];
 	vsY = slice.voxelSize[1];
 	vsZ = slice.voxelSize[2];
 	sliceSpacing = slice.sliceSpacing;
	
 	if (sliceSpacing<0) { 
 		cout << "Slice Spacing not found, assuming 0"<< endl;
		sliceSpacing = 0;//voxelSize[2];
 	}
	resize(slice.dimX,slice.dimY,end-start);
 	setSlice(2,0,slice);
 	for (i=start+1;i<end;i++)  // for all slices
 	{
 		slice.read(file[i]);
		if (reduce) slice.reduceBits(8); // reducing bit size
		if (norm) slice.norm();		// Normalise
		switch (side){
		case 'R': cout << "Cutting Right breast" << endl;
				slice.cut(0,0,(slice.dimX/2)-1,slice.dimY-1);break;
		case 'L': cout << "Cutting Left breast" << endl;
				slice.cut( slice.dimX/2,0,slice.dimX-1,slice.dimY-1);
				break;
		}
		setSlice(2,i-start,slice);  //setting Z0 slice
		//slice.Save("slice.byte");
		//cout << "Press "<<endl;
		//getchar();
 	}
	printInfo();
}

void rmImage3d::readImage(char *baseName)
{
	int i;
 	rmImage slice;
	char arrayQsort[200][200];
	char* point;
	char strNumber[3];
	int maxSlice=-1;
	DIR *dirp = opendir(".");
	struct dirent* dp;
	int notNum;

	for (int i=0;i<200;i++)
	{
		strcpy(arrayQsort[i],"");
	}

		
	while ((dp = readdir(dirp)) != NULL){
		point = strstr(dp->d_name, baseName);
		notNum=false;
    if ((point != NULL) &&
			(strlen(dp->d_name)>=strlen(baseName)+2))
		{
			// getting the number
			strNumber[0]=point[strlen(baseName)];
			strNumber[1]=point[strlen(baseName)+1];
			strNumber[2]=point[strlen(baseName)+2];
			//cout << strNumber<<endl;
			for (int i=0;i<3;i++)
			{
				if (!isdigit(strNumber[i]))
					notNum = true;
			}
			if (!notNum) {
				maxSlice++;
				strcpy(arrayQsort[maxSlice],dp->d_name);
				//cout << "array " << arrayQsort[maxSlice]<<endl;
			} 
		}
	}
	closedir(dirp);
	maxSlice++;
//	cout <<"Sort"<<endl;
//	qsort((void*) arrayQsort,200,200,pstrcmp);
//	cout <<"done"<<endl;
//	for (int j=0;j<maxSlice;j++) {
//			cout << " - "<<arrayQsort[j];
//	}
//	cout<< endl;
	
	slice.read(arrayQsort[0]);
	bits = slice.bits;
	MaxValue = slice.MaxValue;
	cout << "Bits: " << bits << " MaxValue: "<< MaxValue << endl;

 	vsX = slice.voxelSize[0];
 	vsY = slice.voxelSize[1];
 	vsZ = slice.voxelSize[2];
 	sliceSpacing = slice.sliceSpacing;
	
 	if (sliceSpacing<0) { 
 		cout << "Slice Spacing not found, assuming 0"<< endl;
		sliceSpacing = 0;//voxelSize[2];
 	}
	resize(slice.dimX,slice.dimY,maxSlice);
 	setSlice(2,0,slice);
 	for (i=1;i<maxSlice;i++)  // for all slices
 	{
 		slice.read(arrayQsort[i]);
		setSlice(2,i,slice);  //setting Z0 slice
 	}
	printInfo();
}

int pstrcmp (const void *p1, const void *p2)
{
	return strcmp(*(char * const*)p1, *(char * const*)p2);
}
// Saving  the image into Zslices. file1.byte ... fileN.byte, 
void rmImage3d::saveImage (char *file)
{
	int k;
 	rmImage slice;
	char fn[100]; 

	cout << "Saving Image "<<endl;
	printInfo();
	for (k=0;k<dimZ;k++) {
		getSlice(2,k,slice);
		sprintf(fn,"%s%03i.byte",file,k);
		slice.save(fn);
	}
}

/*
///////////
// Rotate 3d Image in different axis by angle phi
//////////
void rmImage3d::rotateX(double phi)
{
	int i,j,k,ni,nj,nk;
	RD_Matrix R(4,4);
	RD_Vector u(4),T;
	rmImage3d result;
	int ndimX,ndimY,ndimZ;
	
	// Rotation matrix in x from [Foley. Computer Graphics]
	R(1,1) = 1; R(1,2) = 0; R(1,3) = 0; R(1,4) = 0;
	R(2,1) = 0; R(2,2) = cos(phi);R(2,3) = -sin(phi); R(2,4) = 0;
	R(3,1) = 0; R(3,2) = sin(phi); R(3,3) = cos(phi); R(3,4) = 0;
	R(4,1) = 0; R(4,2) = 0; R(4,3) = 0; R(4,4) = 1;

 
 	ndimX = 0;
	ndimY = 0;
	ndimZ = 0;

	// Need to compute the new bounding box to resize
	for (i=0;i<dimX;i+=dimX-1){
	  u(1)= i;
	  for (j=0;j<dimY;j+=dimY-1) {
	  	u(2) = j;
	   	for (k=0;k<dimZ;k+=dimZ-1) {
			u(3) = k;
			T = R*u;
			if (ndimX < round(T(1))) ndimX = round(T(1));
			if (ndimY < round(T(2))) ndimY = round(T(2));
			if (ndimZ < round(T(3))) ndimZ = round(T(3));
		}
	  }
	}
		
	cout << "Dimensions " << dimX << " " << dimY << " " << dimZ 
	     << " new " << ndimX << " " << ndimY << " " << ndimZ << endl;
	result.resize(ndimX,ndimY,ndimZ);
	
	// using inverse matrix.
	R = R.inverse();
	
	cout << "Rotation in X:"<< phi << "... ";
    for (i=0;i<result.dimX;i++){
	  u(1)= i;
	  for (j=0;j<result.dimY;j++) {
	  	u(2) = j;
	   	for (k=0;k<result.dimZ;k++) {
			u(3)= k;
			T = R*u;	// Matrix multiplication.
			// Should use some sort of interpolation, just rounding now.
			ni = round(T(1));
			nj = round(T(2));
			nk = round(T(3));
			if ((ni>=0)&&(ni<dimX) && (nj>=0)&&(nj<dimY) 
			  &&(nk>=0)&&(nk<dimZ)) {
				result(i,j,k) = this(ni,nj,nk);
				//cout << "in" <<ni << " " << nj << " " << nk << endl;
			}
		}
	  }
	}
//	this.resize(
	//copyFrom(result);
//	data = result.data;  // Copying result

	resize(result.dimX,result.dimY,result.dimZ);
   	for (i=0;i<dimX;i++)
	  for (j=0;j<dimY;j++) 
	   	for (k=0;k<dimZ;k++) {
		data[i][j][k] = result.data[i][j][k];
	}
	cout << "done."<< endl;
}

void rmImage3d::copyFrom( rmImage3d r)
{
	int i,j,k;
	
	resize(r.dimX,r.dimY,r.dimZ);
	bits = r.bits;
	MaxValue= r.MaxValue;
 	vsX = r.vsX;
 	vsY = r.vsY;
 	vsZ = r.vsZ;
 	sliceSpacing = r.sliceSpacing;
	
	
   	for (i=0;i<dimX;i++)
	  for (j=0;j<dimY;j++) 
	   	for (k=0;k<dimZ;k++) {
		data[i][j][k] = r.data[i][j][k];
	}

}

*/
void rmImage3d::resize(int dx,int dy, int dz, int vx, int vy, int vz)
{
	deleteImage();
	dimX=dx;
	dimY=dy;
	dimZ=dz;
	vsX= vx;
	vsY= vy;
	vsZ= vz;
	allocateImage();
}

void rmImage3d::resize(int dx, int dy, int dz)
{
	deleteImage();
	dimX=dx;
	dimY=dy;
	dimZ=dz;
	allocateImage();
}

void rmImage3d::resize(int *d)
{
	deleteImage();
	MaxValue=255;
	bits=8;
	dimX=d[0];
	dimY=d[1];
	dimZ=d[2];
	vsX= 1;
	vsY= 1;
	vsZ= 1;
	origin[0]=origin[1]=origin[2]=0;
	allocateImage();
}

void rmImage3d::setOrigin(float* org)
{
	origin[0]=org[0];
	origin[1]=org[1];
	origin[2]=org[2];
}

rmImage3d::~rmImage3d()
{
   deleteImage();
}

void rmImage3d::deleteImage()
{
   dimX=dimY=dimZ=0;
}

void rmImage3d::allocateImage()
{
	int i,j,k;
	data.resize(dimX,dimY,dimZ);
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
		  	data(i,j,k)=0;
}

rmImage3d::rmImage3d(char *filename)
{
   readImage(filename);
}

template<class T>
void rmImage3d::copyFrom (const rmMatrix3d<T> &m)
{
	int i,j,k;

	for (k=0;k<dimZ;k++)
       for (j=0;j<dimY;j++)
	      for(i=0;i<dimX;i++)
    	       data(i,j,k) = (imdata3d)m(i,j,k);
}

template<class T>
void rmImage3d::copyTo (rmMatrix3d<T> &m)
{
	int i,j,k;

	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				m(i,j,k) = (T)data(i,j,k);
}

void rmImage3d::copyTo(imdata3d *ptr)
{
	int i,j,k;

	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				ptr[i+dimX*((dimY-1-j)+dimY*k)]= data(i,j,k);

}

void rmImage3d::copyFrom(imdata3d *ptr)
{
	int i,j,k;

	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				data(i,j,k) = ptr[i+dimX*((dimY-1-j)+dimY*k)];

}

void  rmImage3d::operator =(const rmImage3d &im1)
{
	int i,j,k;

	resize(im1.dimX,im1.dimY,im1.dimZ);
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				data(i,j,k) = im1.data(i,j,k);

	bits = im1.bits;
	format = im1.format;
	MaxValue = im1.MaxValue;
	vsX = im1.vsX;
	vsY = im1.vsY;
	vsZ = im1.vsZ;
	sliceSpacing=im1.sliceSpacing;
}

void rmImage3d::operationAnd(rmImage3d im)
{
	int i,j,k;
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				data(i,j,k) = data(i,j,k) & im(i,j,k);
}

void rmImage3d::operationSub(rmImage3d im)
{
	int i,j,k;
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				data(i,j,k) = abs(data(i,j,k) - im(i,j,k));
}

void rmImage3d::invert()
{
	int i,j,k;
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++)
				data(i,j,k) = 255 - data(i,j,k);
}



////////////////////////
/// Get an slice in a particular dimension
//
//   dim: dimension (x:0, y:1, z:2)
//   num: slice number
//   slice: 2d image
///////////////////////
void rmImage3d::getSlice(int dim, int num, rmImage &slice)
{
    int i,j;

    switch (dim){ 
    case 0: 	// Get X slice 
    	slice.resize(dimZ,dimY);
    	for (i=0;i<dimZ;i++){
				for (j=0;j<dimY;j++) {
					if (!slice.outLimit(i,j,0))
						slice.data(i,j)= (imdata)data(num,j,i);
				}
			}
		break;
		case 1:   	// Get Y slice
			slice.resize(dimX,dimZ);
			for (i=0;i<dimX;i++){
				for (j=0;j<dimZ;j++) {
					if (!slice.outLimit(i,j,0))
						slice.data(i,j)= (imdata)data(i,num,j);
				}
			}
			break;
     case 2: 	// Get Z slice
			slice.resize(dimX,dimY);
			for (i=0;i<dimX;i++){
				for (j=0;j<dimY;j++) {
					if (!slice.outLimit(i,j,0))
          	slice.data(i,j)= (imdata)data(i,j,num);
				}
			}
			break;
	}
}
// sets a 2d image as a slice in 3d image.
void rmImage3d::setSlice(int dim, int num, rmImage &slice)
{
    int i,j;

    switch (dim){
    case 0:     // set X slice
			for (i=0;i<dimZ;i++){
				for (j=0;j<dimY;j++) {
					if (!slice.outLimit(i,j,0))
      				data(num,j,i)= (imdata3d)slice.data(i,j);
					else data(num,j,i)=0;
				}
			}
			break;
     case 1:    // set Y slice
        for (i=0;i<dimX;i++){
					for (j=0;j<dimZ;j++) {
						if (!slice.outLimit(i,j,0))
							data(i,num,j)= (imdata3d)slice.data(i,j);
						else data(i,num,j)=0;
					}
				}
        break;
     case 2:    // set Z slice
        for (i=0;i<dimX;i++){
					for (j=0;j<dimY;j++) {
						if (!slice.outLimit(i,j,0))
							data(i,j,num)= (imdata3d)slice.data(i,j);
						else data(i,j,num)=0;
					}
				}
        break;
    }
}

/// reduce image to 8 bits
void rmImage3d::reduceBits()
{
	long no;
	int i,j,k,nb;
	
	nb = 8;
	
	if (bits > nb)
	{
		no=1<<nb;
		cout << "Reduce to "<< no<<endl;
		for(i=0;i<dimX;i++)
		  for (j=0;j<dimY;j++)
	  		for(k=0;k<dimZ;k++){
              	data(i,j,k)=long(data(i,j,k)*no)/MaxValue;
		}
		bits=nb;
		MaxValue=no;
	} else cout <<"reduceBits: image already in 8 bits"<<endl;
}


///////////////////
// Apply Threshold to an image 
///////////////////
void rmImage3d::threshold(int td,int min, int max)
{
	int i,j,k;
 
   
   	for(i=0;i<dimX;i++)
 	  for (j=0;j<dimY;j++)
	     for (k=0;k<dimZ;k++){
			if (data(i,j,k)<td) data(i,j,k)=min;
			else data(i,j,k)=max;
    }
}

void rmImage3d::normalise()
{
	double ratio;
	long range[2];
	int i,j,k;
	
	// Generate histogram
	range[0]=MaxValue;
	range[1]=0;
	for (i=0;i<dimX;i++)
	  for (j=0;j<dimY;j++)
	    for (k=0;k<dimZ;k++) {
		if (data(i,j,k)>range[1])
			range[1]=data(i,j,k);
		if (data(i,j,k)<range[0])
			range[0]=data(i,j,k);
	}
	cout << "Range " << range[0] <<" " << range[1];
	if (range[1]!=range[0]) {
	 	ratio= (MaxValue-1)/double(range[1]-range[0]);
		for (i=0;i<dimX;i++)
	   	 for (j=0;j<dimY;j++)
          for(k=0;k<dimZ;k++) {
                data(i,j,k)=imdata3d((data(i,j,k)-range[0])*ratio);
    	}
	}
//	/*
	range[0]=MaxValue;
	range[1]=0;
	for (i=0;i<dimX;i++)
	  for (j=0;j<dimY;j++)
	    for (k=0;k<dimZ;k++) {
		if (data(i,j,k)>range[1])
			range[1]=data(i,j,k);
		if (data(i,j,k)<range[0])
			range[0]=data(i,j,k);
	}
	cout << "new Range " << range[0] <<" " << range[1]<< endl;
	//*/
}

void rmImage3d::getBBox(rmPoint3d p[2])
{
	int j,i,k;
	p[0].x(dimX);p[0].y(dimY);p[0].z(dimZ);
	p[1].x(0);p[1].y(0);p[1].z(0);
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
    	for(i=0;i<dimX;i++){
				if (data(i,j,k)>0) { 
		    		if (i<p[0].x()) p[0].x(i);
		    		if (j<p[0].y()) p[0].y(j);
		    		if (k<p[0].z()) p[0].z(k);	
						if (i>p[1].x()) p[1].x(i);
						if (j>p[1].y()) p[1].y(j);
						if (k>p[1].z()) p[1].z(k);
				}
		}
}


void rmImage3d::clear(int value)
{
	int j,i,k;
	for (k=0;k<dimZ;k++)
		for (j=0;j<dimY;j++)
			for(i=0;i<dimX;i++) {
		  	data(i,j,k)=value;
	}
}

int rmImage3d::outLimit(int x, int y, int z, int s)
{
	if ((x<s)||(x>=(dimX-s))||
		(y<s)||(y>=(dimY-s))||
		(z<s)||(z>=(dimZ-s))) return true;
	else return false;
}

/////////////////////////////////////
// Enlarge image taking offset x, ofy ,dimX,dimY 
/////////////////////////////////////
void rmImage3d::enlarge(int ofx, int ofy, int ofz,
				int dx, int dy, int dz)
{
	int j,i,k;
	rmImage3d bak;

	bak.resize(dx,dy,dz);
	bak.clear(0);
	
	if (!bak.outLimit(dimX-1,dimY-1,dimZ-1,0)) {
		for (k=0;k<dimZ;k++) {
   		  for (j=0;j<dimY;j++) 
   	    	for(i=0;i<dimX;i++)
       		{
               	bak(i+ofx,j+ofy,k+ofz)=data(i,j,k);
 
       		}
		}
		resize(dx,dy,dz);
		data = bak.data;
	} else cout <<"Enlarge: new image is smaller than original"<< endl;
}

/////////////////////////////////////
// Mirror image in X 
/////////////////////////////////////
void rmImage3d::mirrorX()
{
	int j,i,k;
    rmImage3d bak;

	bak.resize(dimX,dimY,dimZ);
	bak.clear(0);
	
	for (k=0;k<dimZ;k++) 
   	  for (j=0;j<dimY;j++) 
   	   	for(i=0;i<dimX;i++)
    	{
           	bak(dimX-1-i,j,k)=data(i,j,k);
 
	   	}
	data = bak.data;
}

void rmImage3d::mirrorY()
{
	int j,i,k;
    rmImage3d bak;

	bak.resize(dimX,dimY,dimZ);
	bak.clear(0);
	
	for (k=0;k<dimZ;k++) 
   	  for (j=0;j<dimY;j++) 
   	   	for(i=0;i<dimX;i++)
    	{
           	bak(i,dimY-1-j,k)=data(i,j,k);
 
	   	}
	data = bak.data;
}
/////////////////////////////////////
// cutimage image from start to end (included)
/////////////////////////////////////
void rmImage3d::cut(rmPoint3d start,rmPoint3d end)
{
	int j,i,k;
    rmImage3d bak;
		
	if (outLimit(start.x(),start.y(),start.z(),0) 
	  ||outLimit(end.x(),end.y(),end.z(),0))
			cout << "rmImage3d: Can't Cut" <<endl;
	else {		
		bak.resize(end.x()-start.x()+1,end.y()-start.y()+1,
				   end.z()-start.z()+1);
	   	for (k=start.z();k<=end.z();k++)
          for (j=start.y();j<=end.y();j++)
    	    for(i=start.x();i<=end.x();i++)	{
                bak(i-start.x(),j-start.y(),k-start.z())=data(i,j,k);
        }
		resize(end.x()-start.x()+1,end.y()-start.y()+1,
			   end.z()-start.z()+1);
        data=bak.data;
	}
}
void rmImage3d::printInfo()
{
	cout << "Bits: "<< bits << "MaxValue  "<< MaxValue << endl;
	cout << "Image dimensions "<< dimX <<" "<< dimY <<" "<< dimZ << endl;
	cout << "Voxel sizes "<< vsX <<" "<< vsY <<" "<< vsZ 
		 << " ZSpacing :" << sliceSpacing << endl;

	
}

// perform gaussian smoothing in x,y,z
void rmImage3d::gaussSmooth()
{
	int i,j,k;
	rmImage3d im1, work, work2;
	// im1 will be the reduced version of this...
  
  	// Check sizes :
  	if ( (dimX<4) | (dimY<4) | (dimZ<4))
    {
      cerr<<"rmImage3d::gaussSmooth : "<<endl;
      cerr<<"Image is too small"<<endl;
      cerr<<"( "<<dimX<<","<<dimY<<" , "<<dimZ<<")"<<endl;
      abort();
    }
   	im1.resize( int(dimX), int(dimY), int(dimZ) ); // gauss x,y,z
   	work.resize(dimX,dimY,dimZ);  // contains gauss in x
    work2.resize(dimX,dimY,dimZ);// contains gauss in x and y
  
  	// Correlate im2 with a 5 x 1 gaussian filter
  	// ( 1 5 8 5 1 )	[ along x direction ]
    for(k=0;k<dimZ;k++) {
    	for (i=0;i<dimY;i++){
			for (j=2;j<dimX-2;j++) {
				work(j,i,k) = (imdata3d)(( (double) data(j-2,i,k) 
				+ (double) data(j+2,i,k)
				+ 5.0 * data(j-1,i,k) + 5.0 * data(j+1,i,k)
				+ 8.0 * data(j,i,k) ) / 20 );
			}
			// Now deal with edge effects :
			// For 0'th element use ( 8 5 1 ) :
			work(0,i,k) = (imdata3d) (( 8.0 * data(0,i,k)
			      + 5.0 * data(1,i,k) + (double) data(2,i,k) ) / 14);
	
			// For 1st element use ( 5 8 5 1 ) :
			work(1,i,k) = (imdata3d) (( 5.0 * data(0,i,k) 
					+ 8.0 * data(1,i,k) + 5.0 * data(2,i,k)
					+ (double) data(3,i,k) ) / 19);
	
			// For last element use ( 1 5 8 ) :
			work(dimX-1,i,k) = (imdata3d) (( 8.0 * data(dimX-1,i,k) 
				 	+ 5.0 * data(dimX-2,i,k) 
				 	+ (double) data(dimX/3,i,k) ) / 14);
	
			// For last but one use ( 1 5 8 5 ) :
			work(dimX-2,i,k) = (imdata3d) (( 5.0 * data(dimX-1,i,k)
				 + 8.0 * data(dimX-2,i,k) 
				 + 5.0 * data(dimX-3,i,k)
				 + data(dimX-4,i,k) ) / 19);
      }
  }
  
  // Correlate work image with 1 x 5 gaussian
  // filter (1 5 8 5 1)
  for(k=0;k<dimZ;k++) {
  	for (j=0;j<dimX;j++){
		for (i=2;i<dimY-2;i++)
	  	{
	    	work2(int(j),int(i),k) = (imdata3d )
				(( (double) work(j,i-2,k) + (double) work(j,i+2,k)
				+ 5.0 * work(j,i-1,k) + 5.0 * work(j,i+1,k)
				+ 8.0  * work(j,i,k) ) / 20);
	  	}

		// Now edge effects :
		// Use (8 5 1) filter for 0th element.
		work2(int(j),0,k) = (imdata3d) ( ( 8.0 * work(j,0,k)
					     + 5.0 * work(j,1,k)
					     + (double) work(j,2,k) ) / 14 );
		// For 1st element use ( 5 8 5 1 ) :
		work2(j,1,k) = (imdata3d) (( 5.0 * work(j,0,k)
					+ 8.0 * work(j,1,k) + 5.0 * work(j,2,k)
					+ (double) work(j,3,k) ) / 19);
	
		// Last element (1 5 8) filter at end.
  		work2(int(j),int(dimY)-1,k)
	    		= (imdata3d) ( ( 8.0 * work(j,dimY-1,k)
		      	+ 5.0 * work(j,dimY-2,k)
		      	+ (double) work(j,dimY-3,k) )/14);

		// Last but one element (1 5 8 5) filter at end.
		work2(int(j),int(dimY)-2,k)
	    		= (imdata3d) (( 5.0 * work(j,dimY-1,k)
		     	+ 8.0 * work(j,dimY-2,k) 
		     	+ 5.0 * work(j,dimY-3,k)
		     	+ (double) work(j,dimY-4,k) )/19 );
  	}
  }
  
  // finally perform filtering in z direction.
  for(i=0;i<dimY;i++) {
  	for(j=0;j<dimX;j++) {
    	for(k=2;k<dimZ-2;k++)
		{
	  		im1(j,i,k)= (imdata3d) (((double)work2(j,i,k-2) +
				(double) work2(j,i,k+2) 
	       		+ 5.0*work2(j,i,k-1) + 5.0*work2(j,i,k+1)
	       		+ 8.0*work2(j,i,k) ) /20);
		}
    
      	// Now deal with edge effects :
      	//Use (8 5 1) filter for 0th element.
      	im1(j,i,0) = (imdata3d)(( 8.0*work2(j,i,0) 
				+ 5.0*work2(j,i,1)
				+ (double)work2(j,i,2)) /14 );
				
		// For 1st element use ( 5 8 5 1 ) :
		im1(j,i,1) = (imdata3d) (( 5.0 * work2(j,i,0) 
					+ 8.0 * work2(j,i,1) + 5.0 * work2(j,i,2)
					+ (double) work2(j,i,3) ) / 19);

      
		// Last element (1 5 8) filter at end.
		im1(j,i,int(dimZ)-1) = 
				(imdata3d)(( 8.0*work2(j,i,dimZ-1)
				+ 5.0*work2(j,i,dimZ-2)
			    + (double) work2(j,i,dimZ-3)) / 14);
				
		// Last but one element (1 5 8 5)
		im1(j,i,int(dimZ)-2) =  
				(imdata3d) ((5.0*work2(j,i,dimZ-1) 
					    + 8.0*work2(j,i,dimZ-2) +
					    5.0*work2(j,i,dimZ-3) + 
					    (double) work2(j,i,dimZ-4)) /19);
	}
  }
  
  data = im1.data;
  dimX = im1.dimX;
  dimY = im1.dimY;
  dimZ = im1.dimZ;
}



// perform gaussian reduction in x,y,z
void rmImage3d::gaussReduce()
{
	int i,j,k;
	rmImage3d im1, work, work2;
	// im1 will be the reduced version of this...
  
  	// Check sizes :
  	if ( (dimX<4) | (dimY<4) | (dimZ<4))
    {
      cerr<<"rmImage3d::gaussReduce : "<<endl;
      cerr<<"Image is too small"<<endl;
      cerr<<"( "<<dimX<<","<<dimY<<" , "<<dimZ<<")"<<endl;
      abort();
    }
   	im1.resize( int(dimX/2), int(dimY/2), int(dimZ/2) ); // gauss x,y,z
   	work.resize(dimX,dimY,dimZ);  // contains gauss in x
    work2.resize(dimX,dimY,dimZ);// contains gauss in x and y
  
  	// Correlate im2 with a 5 x 1 gaussian filter
  	// ( 1 5 8 5 1 )	[ along x direction ]
    for(k=0;k<dimZ;k++) {
    	for (i=0;i<dimY;i++){
			for (j=2;j<dimX-2;j++) {
				work(j,i,k) = (imdata3d)(( (double) data(j-2,i,k) 
				+ (double) data(j+2,i,k)
				+ 5.0 * data(j-1,i,k) + 5.0 * data(j+1,i,k)
				+ 8.0 * data(j,i,k) ) / 20 );
			}
			// Now deal with edge effects :
			// For 0'th element use ( 8 5 1 ) :
			work(0,i,k) = (imdata3d) (( 8.0 * data(0,i,k)

			      + 5.0 * data(1,i,k) + (double) data(2,i,k) ) / 14);
	
			// For 1st element use ( 5 8 5 1 ) :
			work(1,i,k) = (imdata3d) (( 5.0 * data(0,i,k) 
					+ 8.0 * data(1,i,k) + 5.0 * data(2,i,k)
					+ (double) data(3,i,k) ) / 19);
	
			// For last element use ( 1 5 8 ) :
			work(dimX-1,i,k) = (imdata3d) (( 8.0 * data(dimX-1,i,k) 
				 	+ 5.0 * data(dimX-2,i,k) 
				 	+ (double) data(dimX/3,i,k) ) / 14);
	
			// For last but one use ( 1 5 8 5 ) :
			work(dimX-2,i,k) = (imdata3d) (( 5.0 * data(dimX-1,i,k)
				 + 8.0 * data(dimX-2,i,k) 
				 + 5.0 * data(dimX-3,i,k)
				 + data(dimX-4,i,k) ) / 19);
      }
  }
  
  // Correlate and subsample work image with 1 x 5 gaussian
  // filter (1 5 8 5 1)
  for(k=0;k<dimZ;k++) {
  	for (j=0;j<(dimX-1);j+=2){
		for (i=2;i<dimY-2;i+=2)
	  	{
	    	work2(int(j/2),int(i/2),k) = (imdata3d )
				(( (double) work(j,i-2,k) + (double) work(j,i+2,k)
				+ 5.0 * work(j,i-1,k) + 5.0 * work(j,i+1,k)
				+ 8.0  * work(j,i,k) ) / 20);
	  	}


		// Now edge effects :
		// Use (8 5 1) filter for 0th element.
		work2(int(j/2),0,k) = (imdata3d) ( ( 8.0 * work(j,0,k)
					     + 5.0 * work(j,1,k)
					     + (double) work(j,2,k) ) / 14 );
	
		// If ny is odd, use (1 5 8 5) filter at end,
		// if ny is even, use (1 5 8) filter at end.
		if (dimY%2 == 0)
	  		work2(int(j/2),int(dimY/2)-1,k)
	    		= (imdata3d) ( ( 8.0 * work(j,dimY-1,k)
		      	+ 5.0 * work(j,dimY-2,k)
		      	+ (double) work(j,dimY-3,k) )/14);
		else
			work2(int(j/2),int(dimY/2)-1,k)
	    		= (imdata3d) (( 5.0 * work(j,dimY-1,k)
		     	+ 8.0 * work(j,dimY-2,k) 
		     	+ 5.0 * work(j,dimY-3,k)
		     	+ (double) work(j,dimY-4,k) )/19 );
  	}
  }
  
  // finally perform filtering & subsampling in z direction.
  for(i=0;i<dimY/2;i++) {
  	for(j=0;j<dimX/2;j++) {
    	for(k=2;k<dimZ-2;k+=2)
		{
	  		im1(j,i,k/2)= (imdata3d) (((double)work2(j,i,k-2) +
				(double) work2(j,i,k+2) 
	       		+ 5.0*work2(j,i,k-1) + 5.0*work2(j,i,k+1)
	       		+ 8.0*work2(j,i,k) ) /20);
		}
    
      	// Now deal with edge effects :
      	//Use (8 5 1) filter for 0th element.
      	im1(j,i,0) = (imdata3d)(( 8.0*work2(j,i,0) 
				+ 5.0*work2(j,i,1)
				+ (double)work2(j,i,2)) /14 );
      
      	// same scheme for odd & even nz as used for ny above.
        if(dimZ%2 == 0)
			im1(j,i,int(dimZ/2)-1) = 
				(imdata3d)(( 8.0*work2(j,i,dimZ-1)
				+ 5.0*work2(j,i,dimZ-2)
			    + (double) work2(j,i,dimZ-3)) / 14);
      	else
			im1(j,i,int(dimZ/2)-1) =  
				(imdata3d) ((5.0*work2(j,i,dimZ-1) 
					    + 8.0*work2(j,i,dimZ-2) +
					    5.0*work2(j,i,dimZ-3) + 
					    (double) work2(j,i,dimZ-4)) /19);
	}
  }
  
  data = im1.data;
  dimX = im1.dimX;
  dimY = im1.dimY;
  dimZ = im1.dimZ;
}



/*
/////////////////////////////////////
// Transform an image by a teta rotation and a ux,uy translation
/////////////////////////////////////
void Image3d::Transf(double teta, double ux,double uy)
{
	int j,i,k,l;
	imdata **pbak,**pdata=data.dataPtr();
	rmImage bak;
	RD_Matrix m(3,3),n(3,1);

	bak.Resize(dimX,dimY);
	pbak=bak.data.dataPtr();	

	m(1,1)=cos(teta); m(1,2)=sin(teta); m(1,3)= ux;
	m(2,1)=-sin(teta); m(2,2)=cos(teta); m(2,3)=uy;
	m(3,1)=0;m(3,2)=0;m(3,3)=1;
	m=m.inverse();
	for (j=0;j<dimY;j++)
           for(i=0;i<dimX;i++)
           {
		n(1,1)=i; n(2,1)=j;n(3,1)=1;
		n=m*n;
		k=int(n(1,1));l=int(n(2,1));	
                if ((k>=0)&&(k<dimX)&&(l>=0)&&(l<dimY))
                        pbak[j][i]=pdata[l][k];
	    }
	data=bak.data;
}

/////////////////////////////////////
// Negate image greylevels 
/////////////////////////////////////
void rmImage::Negate()
{
        int j,i;

        for (j=0;j<dimY;j++)
           for(i=0;i<dimX;i++)
           {
           }
}

/////////////////////////////////////
// Translate an image ux,uy 
/////////////////////////////////////
void rmImage::Translate(double ux,double uy)
{
	int j,i,k,l;
        imdata **pbak,**pdata=data.dataPtr();
        rmImage bak;
 
        bak.Resize(dimX,dimY);
        pbak=bak.data.dataPtr();
        for (j=0;j<dimY;j++)
           for(i=0;i<dimX;i++)
           {
                k=int(i+ux);l=int(j+uy);
                if ((k>=0)&&(k<dimX)&&(l>=0)&&(l<dimY))
                        pbak[j][i]=pdata[l][k];
            }
        data=bak.data;
}

/////////////////////////////////
// Add Gaussian noise to the image
///////////////////
void rmImage::Noise(float m,float std)
{	
	int i,j;
	SM_Random rand;
	int value;
        for (j=0;j<dimY;j++)
         for(i=0;i<dimX;i++)
         {
	    	value = int(rand.gaussian(m,std));
		if ((ptr[j][i]+value)>=MaxValue) ptr[j][i]=MaxValue-1;
		else if ((ptr[j][i]+value)<0)  ptr[j][i]=0; 
        	     else ptr[j][i]=ptr[j][i]+value;
         }
}

///////////////////
// Apply Threshold to an image 
///////////////////
void rmImage::Threshold(int td,int min, int max)
{
        int i,j;
 
        for (j=0;j<dimY;j++)
         for(i=0;i<dimX;i++)
         {
		//if (ptr[j][i]<td) ptr[j][i]=min;
		//else ptr[j][i]=max;
         }
}

///////////////////
// Return the number of different grey-levels of the image 
///////////////////
int rmImage::NumGrey()
{
        int i;
	Histo1d h(MaxValue);
	int num=0;

	h.Create(*this);
        for (i=0;i<MaxValue;i++)
           if(h.ReadValue(i)>0) num++;
	return num;
}



///////////////////////////
// Read/Write routines used in den files...
////////////////////////////
int Image3d::read_bytes(int fd, char *buf, int bytecount)
{
    int n;

    while (bytecount > 0) {
        n = MIN(bytecount, MAX_READ_SIZE);
        if (read(fd, buf, n) != n)
            return(0);
        buf += n;
        bytecount -= n;
    }
    return(1);
}

int Image3d::read_shorts(int fd, short *sbuf, int shortcount, int swap)
{
    int n, c;
    int bytecount = shortcount * 2;
    char tmp0, tmp1, tmp2, tmp3;
    char *buf = (char *)sbuf;

    while (bytecount > 0) {
        n = MIN(bytecount, MAX_READ_SIZE);
        if (read(fd, buf, n) != n)
            return(0);
        bytecount -= n;
        if (swap) {
            c = n / 8;
            n -= c * 8;
            for (; c > 0; c--) {
                tmp0 = buf[0]; buf[0] = buf[1]; buf[1] = tmp0;
                tmp1 = buf[2]; buf[2] = buf[3]; buf[3] = tmp1;
                tmp2 = buf[4]; buf[4] = buf[5]; buf[5] = tmp2;
                tmp3 = buf[6]; buf[6] = buf[7]; buf[7] = tmp3;
                buf += 8;
            }
            for (; n > 0; n -= 2) {
                tmp0 = buf[0]; buf[0] = buf[1]; buf[1] = tmp0;
                buf += 2;
            }
        } else {
            buf += n;
        }
    }
    return(1);
}

int Image3d::read_words(int fd, int *wbuf, int wordcount, int swap)
{
    int n, c;
    int bytecount = wordcount * 4;
    char tmp0, tmp1, tmp2, tmp3;
    char *buf = (char *)wbuf;

    while (bytecount > 0) {
        n = MIN(bytecount, MAX_READ_SIZE);
        if (read(fd, buf, n) != n)
            return(0);
        bytecount -= n;
        if (swap) {
            c = n / 8;
            n -= c * 8;
            for (; c > 0; c--) {
                tmp0 = buf[0]; buf[0] = buf[3]; buf[3] = tmp0;
                tmp1 = buf[1]; buf[1] = buf[2]; buf[2] = tmp1;
                tmp2 = buf[4]; buf[4] = buf[7]; buf[7] = tmp2;
                tmp3 = buf[5]; buf[5] = buf[6]; buf[6] = tmp3;
                buf += 8;
            }
            for (; n > 0; n -= 4) {
                tmp0 = buf[0]; buf[0] = buf[3]; buf[3] = tmp0;
                tmp1 = buf[1]; buf[1] = buf[2]; buf[2] = tmp1;
                buf += 4;
            }
        } else {
            buf += n;
        }
    }
    return(1);
}

int Image3d::write_bytes(int fd, char *buf, int bytecount)
{
    int n;

    while (bytecount > 0) {
        n = MIN(bytecount, MAX_READ_SIZE);
        if (write(fd, buf, n) != n)
            return(0);
        buf += n;
        bytecount -= n;
    }
    return(1);
}

*/
