// 3D Matrix Class
//    Robert Marti
//	May 2001 

/* Define as we already include the implementation in the header (to avoid
redefinition*/
#ifndef RMMATRIX3D_C
#define RMMATRIX3D_C

#include <libgen.h>
#include <stdio.h>
#include "rmmatrix3d.h"

template<class T>
rmMatrix3d<T>::rmMatrix3d()
{
	data = NULL;
	dimX=dimY=dimZ=0;
	orgX=orgY=orgZ=0;
}

template<class T>
rmMatrix3d<T>::rmMatrix3d(int dx,int dy, int dz)
{
	data = NULL;
	dimX=dimY=dimZ=0;
	orgX=orgY=orgZ=0;
	resize(0,0,0,dx,dy,dz);
}

template<class T>
rmMatrix3d<T>::rmMatrix3d(int ox,int oy, int oz,int dx,int dy, int dz)
{
	data = NULL;
	dimX=dimY=dimZ=0;
	orgX=orgY=orgZ=0;
	resize(ox,oy,oz,dx,dy,dz);
}

// Copy constructor
template<class T>
rmMatrix3d<T>::rmMatrix3d<T>(const rmMatrix3d<T> &im1)
{
	int i,j,k;

	resize(im1.orgX,im1.orgY,im1.orgZ,im1.dimX,im1.dimY,im1.dimZ);
	for (k=orgZ;k<dimZ;k++)
		for (j=orgY;j<dimY;j++)
			for(i=orgX;i<dimX;i++)
				data[(i-orgX)+(dimX-orgX)*((j-orgY)+(dimY-orgY)*(k-orgZ))]=
				im1(i,j,k);

}


template<class T>
const T & rmMatrix3d<T>::operator() (int x, int y, int z) const
{
	if ((x<orgX) || (x>=dimX) || (y<orgY) || (y>=dimY)
			||(z<orgZ) || (z>=dimZ))
	{
		cerr<<"Error rmMatrix3d: Out of bounds "<<x << " "<< y<< " " <<z <<endl;
		exit(-1);
	}
	else
	return data[(x-orgX)+(dimX-orgX)*((y-orgY)+(dimY-orgY)*(z-orgZ))];
}

template<class T>
T & rmMatrix3d<T>::operator() (int x, int y, int z)
{
	if ((x<orgX) || (x>=dimX) || (y<orgY) || (y>=dimY)
			||(z<orgZ) || (z>=dimZ))
	{
		cerr<<"Error rmMatrix3d: Out of bounds "<<x << " "<< y<< " " <<z <<endl;
		exit(-1);
	}
	return data[(x-orgX)+(dimX-orgX)*((y-orgY)+(dimY-orgY)*(z-orgZ))];
}

template<class T>
void rmMatrix3d<T>::resize(int dx, int dy, int dz)
{
	resize(0,0,0,dx,dy,dz);
}

template<class T>
void rmMatrix3d<T>::resize(int ox,int oy, int oz,int dx,int dy, int dz)
{
    deleteMatrix();
	orgX=ox;
	orgY=oy;
	orgZ=oz;
    dimX=dx;
	dimY=dy;
	dimZ=dz;
	allocateMatrix();
}

template<class T>
void rmMatrix3d<T>::resize(int *d)
{
	resize(0,0,0,d[0],d[1],d[2]);
}


template<class T>
rmMatrix3d<T>::~rmMatrix3d()
{
   deleteMatrix();
}

template<class T>
void rmMatrix3d<T>::deleteMatrix()
{
   delete[] data;
   dimX=dimY=dimZ=0;
   orgX=orgY=orgZ=0;
}

template<class T>
void rmMatrix3d<T>::allocateMatrix()
{
   data = new T [(dimX-orgX)*(dimY-orgY)*(dimZ-orgZ)];
}

template<class T>
void  rmMatrix3d<T>::operator =(const rmMatrix3d<T> &im1)
{
	int i,j,k;

	resize(im1.orgX,im1.orgY,im1.orgZ,im1.dimX,im1.dimY,im1.dimZ);
	for (k=orgZ;k<dimZ;k++)
       for (j=orgY;j<dimY;j++)
	      for(i=orgX;i<dimX;i++)
	  		   data[(i-orgX)+(dimX-orgX)*((j-orgY)+(dimY-orgY)*(k-orgZ))]=
			   		im1(i,j,k);

}

template<class T>
void rmMatrix3d<T>::printInfo()
{
	cout << "Matrix dimensions: origin "<< 
			orgX <<" "<< orgY <<" "<< orgZ << " dim " <<
			dimX <<" "<< dimY <<" "<< dimZ << endl;
}
#endif
