/***************************************************************************
                          rmrmComat.cpp  -  description
                             -------------------
    begin                : Thu Jul 4 2002
    copyright            : (C) 2002 by Robert Marti
    email                : robertmarti@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include "rmcomat.h"
#include <iostream.h>

rmComat::rmComat()
{
	N=0;
}

rmComat::rmComat(int n)
{
	init(n);
}

void rmComat::init(int n)
{
	int i,j,k;

	N=n;

	ca = new double *[N];
	cb = new double *[N];
 	for (i=0;i<N;i++)
	{	
		ca[i]=new double[N];
		cb[i]=new double[N];
	}	

	c= new double ***[N];
	if (c==NULL) cout<< "Memory assignement: c"<<endl;
	
	for (i=0;i<N;i++)
	{
	  c[i]= new double **[N];
	  if (c[i]==NULL) cerr<< "Memory assignement: c[i]"<<endl;
	  for (j=0;j<N;j++)
	  {
	    c[i][j]= new double *[N];
	    if (c[i][j]==NULL) cerr<< "Memory assignement: c[i][j] c"<<endl;
	    for (k=0;k<N;k++)
	    {
				c[i][j][k]=new double[N];
				if (c[i][j][k]==NULL)cerr<<"Memory assignement: c[i][j][k]"<<endl;
	    }
	  }
	}
	div=256/N;
	reset();
}

void rmComat::reset()
{
	int i,j,k,l;

	NP=0;
	for (i=0;i<N;i++)
		for (j=0;j<N;j++)
			for (k=0;k<N;k++)
				for (l=0;l<N;l++){
					c[i][j][k][l]=0;
	}
}

rmComat::~rmComat()
{
	int i,j,k;

	for (i=0;i<N;i++)
	{
		for (j=0;j<N;j++)
		{
			for (k=0;k<N;k++)
			{
				delete c[i][j][k];
			}
	    delete c[i][j];
		}
		delete c[i];
	  delete ca[i];
	  delete cb[i];
	}
	delete c;
}

void rmComat::calc(const rmImage &imA, const rmImage &imB, int t, double a)
{
	int x,y,i,j;

	angle=a;
	trans=t;

	//Reset();	
	x=int(trans*cos(a));
	y=int(trans*sin(a));
	int dx,dy;
	if (imA.dimX>imB.dimX) dx = imB.dimX;
	else dx =imA.dimX;
	if (imA.dimY>imB.dimY) dy = imB.dimY;
	else dy =imB.dimY;
	for (i=0;i<dx;i++)
		for (j=0;j<dy;j++)
	  {
			if (((i+x)>=0)&&((i+x)<dx)&& ((j+y)>=0)&&((j+y)<dy))
			{
				(c[imA.data(i,j)/div] [imA.data(i,j)/div]
					[imB.data(i,j)/div] [imB.data(i,j)/div])++;
				NP++;
			}
	  }
	normalise();
	calcMarginal();
}

// Normalise 4d GLCM structure
void rmComat::normalise()
{
	int i,j,k,l;

	 for (i=0;i<N;i++)
		for (j=0;j<N;j++)
			for (k=0;k<N;k++)
	      for (l=0;l<N;l++) {
					c[i][j][k][l]=c[i][j][k][l]/NP;
	}
}

// Calculate marginal distributions
void rmComat::calcMarginal()
{
	int i,j,k,l;

	for (i=0;i<N;i++)
		for (j=0;j<N;j++) {
	    ca[i][j]=0;
	    cb[i][j]=0;
			for (k=0;k<N;k++)
				for (l=0;l<N;l++) {
					ca[i][j]+=c[i][j][k][l];
					cb[i][j]+=c[k][l][i][j];
			}
	}
}

// Add a value to the GLCM matrix
void rmComat::add(int a1, int a2,int a3,int a4,double value)
{
	c[a1/div][a2/div][a3/div][a4/div]+=value;
	NP+=value;
}
