#include "rmutils.h"
#include <math.h>
#include <stdlib.h>
//#include "NumRec/Gamma.h"
#define TINY 1.0e-20
#define MAXIT 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30

void error (char *str)
{
	//cout << "Error" << str;
	//getch();
	//perror(str);
	cerr << "Error "<< str;
	exit(-1);
}

// Linear correlation of set of 2d points
template<class T>
void pearsn (rmVector <T> x, rmVector<T> y, unsigned long n, float &r,float &prob,float &z)
{
	float betai (float a,float b, float c);
	float erfcc(float x);
	unsigned long j;
	float yt,xt,t,df;
	float syy=0.0,sxy=0.0,sxx=0.0,ay=0.0,ax=0.0;
	
	for (j=1;j<=n;j++)
	{
		ax+= x(j);
		ay+= y(j);
	}
	ax /= n;
	ay /= n;
	for (j=1;j<=n;j++) 
        { 
		xt=x(j)-ax;
		yt=y(j)-ay;
		sxx+= xt*xt;
		syy+= yt*yt;
		sxy+= xt*yt;
	}
	r=sxy/sqrt(sxx*syy);
	z=0.5*log((1.0+r+TINY)/(1.0-r+TINY));
	df = n-2;
	t=r*sqrt(df/((1.0-r+TINY)*(1.0+r+TINY)));
	//prob=betai(0.5*df,0.5,df/(df+t*t));
	prob = erfcc(fabs(z*sqrt(n-1.0))/1.4142136);
}


// Returns the incomplete beta function Ix(a,b)
float betai (float a, float b, float x)
{
	float betacf(float a,float b, float x);
	float bt;

/*
	if (x<0.0 || x>1.0) error("Bad x in routine betai");
	if (x==0.0 || x==1.0) bt = 0.0;
	else
		bt=exp(NR_lnGamma(a+b)-NR_lnGamma(a)-NR_lnGamma(b)+
			a*log(x)+b*log(1.0-x));
	if (x< (a+1.0)/(a+b+2.0))
		return bt*betacf(a,b,x)/a;
	else 
		return 1.0-bt*betacf(b,a,1.0-x)/b;
 */
	cout << "To do lnGamma from numrec"<< endl;
	return 1;
}

float betacf (float a, float b, float x)
{

	int m,m2;
	float aa,c,d,del,h,qab,qam,qap;

	qab=a+b;
	qap=a+1.0;
	qam=a-1.0;
	c=1.0;
	d=1.0-qab*x/qap;
	if (fabs(d) < FPMIN) d=FPMIN;
	d=1.0/d;
	h=d;
	for (m=1;m<=MAXIT;m++)
	{
		m2=2*m;
		aa=m*(b-m)*x/((qam+m2)*(a+m2));
		d=1.0+aa*d;
		if (fabs(d)<FPMIN) d=FPMIN;
		c=1.0+aa*c;
		if (fabs(c)<FPMIN) c=FPMIN;
		d=1.0/d;
		h *= d*c; 
		aa = -(a+m)*(qab+m)*x/((a+m2)*(qap+m2));
		d=1.0+aa*d;
		if (fabs(d)<FPMIN) d=FPMIN;
		c=1.0+aa*c; 
                if (fabs(c)<FPMIN) c=FPMIN;  
                d=1.0/d;
		del=d*c;
		h*=del;
		if (fabs(del-1.0)<EPS) break;
	}
	if (m> MAXIT) error ("a or b is too big or MAXIT too small in betacf");
	return h;
}

float erffc(float x)
{
	//return x <0.0 ? 1.0+NR_GammaP(0.5,x*x) : NR_GammaQ(0.5,x*x);
	cout << "To do lnGamma from numrec"<< endl;
	return 1;
}

float erfcc(float x)
{
	float t,z,ans;

	z=fabs(x);
	t=1.0/(1.0+0.5*z);
	ans = t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+
		t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
		t*(-0.82215223+t*0.17087277)))))))));
	return x >= 0.0 ? ans : 2.0-ans;
}


template<class T>
void moment ( rmVector<T> data, int n, float &ave, float &adev, float &sdev,
        float &var, float &skew, float &curt)
{
	int j;
	float ep=0.0,s,p;

	if (n<=1) error("n must be at least 2 in moment");
	s=0.0;
	for (j=1;j<=n;j++) s += data(j);
	ave=s/n;
	adev=var=skew=curt=0.0;
	for (j=1;j<=n;j++)
	{
		adev += fabs(s=data(j)-ave);
		var += (p=s*s);
		skew += (p*=s);
		curt += (p*=s);
	}
	adev /=n;
	var = (var-ep*ep/n)/(n-1);
	sdev = sqrt(var);
	if (var) 
	{
		skew/= (n*var*sdev);
		curt= curt/(n*var*var)-3.0;
	} else error("No skew/kurtosis when variance = 0 in moment");
}

void moment ( long *data, int n, float &ave, float &adev, float &sdev, 
        float &var, float &skew, float &curt)
{
	int j;
	float ep=0.0,s,p;

	if (n<=1) error("n must be at least 2 in moment");
	s=0.0;
	for (j=1;j<=n;j++) s += data[j];
	ave=s/n;
	adev=var=skew=curt=0.0;
	for (j=1;j<=n;j++)
	{
		adev += fabs(s=data[j]-ave);
		var += (p=s*s);
		skew += (p*=s);
		curt += (p*=s);
	}
	adev /=n;
	var = (var-ep*ep/n)/(n-1);
	sdev = sqrt(var);
	if (var) 
	{
		skew/= (n*var*sdev);
		curt= curt/(n*var*var)-3.0;
	} else error("No skew/kurtosis when variance = 0 in moment");
}

/*
void moment ( rmList<int> data, int n, double &ave, double &sdev)
{
	int j;
	double ep=0.0,s,p,var,adev;

	if (n<=1) error("n must be at least 2 in moment");
	s=0.0;
	for (j=0;j<n;j++) s += data(j);
	ave=s/n;
	var=adev=0.0;
	for (j=0;j<n;j++)
	{
		//cout << "data "<< j << " "<< data(j)<< endl;
		adev += fabs(s=double(data(j))-ave);
		var += (p=s*s);
	}
	adev /=n;
	var = (var-ep*ep/n)/(n-1);
	sdev = sqrt(var);
	//cout << "Mean "<< ave << " stdev" << sdev;
}
*/
template<class T>
void moment ( rmVector<T> data, int n, double &ave, double &sdev)
{
	int j;
	double ep=0.0,s,p,var,adev;

	if (n<=1) error("n must be at least 2 in moment");
	s=0.0;
	for (j=1;j<=n;j++) s += data(j);
	ave=s/n;
	adev=var=0.0;
	for (j=1;j<=n;j++)
	{
		adev += fabs(s=data(j)-ave);
		var += (p=s*s);
	}
	adev /=n;
	var = (var-ep*ep/n)/(n-1);
	sdev = sqrt(var);
}

// Euclidean distance between two points
double dist(double p1x,double p1y,double p2x,double p2y,double &dx,double &dy)
{
	dx = fabs(p1x-p2x);
	dy = fabs(p1y-p2y);
	return (sqrt(dx*dx+dy*dy));
}


// Euclidean distance between two 3d  points 
double dist3d(double p1x,double p1y,double p1z ,
			double p2x,double p2y,double p2z,
			double &dx,double &dy,double &dz)
{
	dx = fabs(p1x-p2x);
	dy = fabs(p1y-p2y);
	dz = fabs(p1z-p2z);
	return (sqrt(dx*dx+dy*dy+dz*dz));
}


// Maximum value 
int MaxVal(int x,int y) {
  if (x>y) return x;
  else return y;
}

// Minimum value
int MinVal(int x,int y) {
  if (x<y) return x;
  else return y;
}

